import aiohttp
from pyroyale.endpoints import API_PATHS
from pyroyale.exceptions import (
    BadRequest,
    Forbidden,
    NotFound,
    UnknownRequest,
    ServiceUnavailable,
    BadMethod,
    Removed,
)


class Route:
    BASE_URL = "https://api.clashroyale.com/v1"

    def __init__(self, method: str, path: str, params: dict = None):
        if "#" in path:
            path = path.replace("#", "%23")

        self.method = method
        self.path = path
        self.params = params
        url = self.BASE_URL + self.path
        self.url = url


class ClashRoyaleAPI:
    """Represents an HTTP Client sending HTTP requests to the Official Clash Royale API.


    Parameters
    ----------
    token : str
        The token provided from the Clash Royale API. https://developer.clashroyale.com/
    """

    def __init__(self, token: str):
        self.token = token

    async def request_validator(self, response: aiohttp.ClientResponse):
        data = await response.text()

        if 300 > response.status >= 200:
            return data
        if response.status == 400:
            raise BadRequest(response, data)
        if response.status == 403:
            raise Forbidden(response, data)
        if response.status == 404:
            raise NotFound(response, data)
        if response.status == 410:
            raise Removed(response, data)
        if response.status == 500:
            raise UnknownRequest(response, data)
        if response.status == 503:
            raise ServiceUnavailable(response, data)

    async def request(self, route):
        method = route.method
        url = route.url
        headers = {"Authorization": f"Bearer {self.token}"}
        parameters = route.params

        # api only utilizes GET requests at the moment, but keep this incase some new endpoints might need it.
        async with aiohttp.ClientSession() as session:
            if method.upper() == "GET":
                async with session.get(url, headers=headers, params=parameters) as resp:
                    return await self.request_validator(resp)
            elif method.upper() == "POST":
                async with session.post(url, headers=headers, params=parameters) as resp:
                    return await self.request_validator(resp)
            else:
                raise BadMethod("ERROR", "Invalid Request method given! GET or POST only!")

    def remove_none(self, params: dict) -> dict:
        return {k: v for k, v in params.items() if v is not None}

    # Clan Information
    async def get_clan_war_log(self, clanTag: str, limit: int = None, after: str = None, before: str = None):
        """
        get_clan_war_log - Retrieve clan's clan war log

        :param clanTag: Tag of the clan
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["retrieve_clan_war_log"].format(clanTag=clanTag),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def search_clans(
        self,
        name: str = None,
        locationId: int = None,
        minMembers: int = None,
        maxMembers: int = None,
        minScore: int = None,
        limit: int = None,
        after: str = None,
        before: str = None,
    ):
        """
        search_clans - Search all clans by name and/or filtering the results using various criteria.
                        At least one filtering criteria must be defined and if name is used as part of search,
                        it is required to be at least three characters long It is not possible to specify ordering
                        for results so clients should not rely on any specific ordering as that may change in the future releases of the API.

        :param name: Search clans by name. If name is used as part of search query, it needs to be at least three characters long.
                    Name search parameter is interpreted as wild card search, so it may appear anywhere in the clan name.
        :param locationId: Filter by clan location identifier. For list of available locations, refer to getLocations operation.
        :param minMembers: Filter by minimum number of clan members
        :param maxMembers: Filter by maximum number of clan members
        :param minScore: Filter by minimum amount of clan score.
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {
            "name": name,
            "locationId": locationId,
            "minMembers": minMembers,
            "maxMembers": maxMembers,
            "minScore": minScore,
            "limit": limit,
            "after": after,
            "before": before,
        }
        r = Route("GET", API_PATHS["search_clans"], params=self.remove_none(query_params))
        return await self.request(r)

    async def get_clan_river_race_log(self, clanTag: str, limit: int = None, after: str = None, before: str = None):
        """
        get_clan_river_race_log - Retrieve clan's river race log

        :param clanTag: Tag of the clan
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["retrieve_clan_river_race_log"].format(clanTag=clanTag),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def get_current_clan_war(self, clanTag: str):
        """
        get_current_clan_war - Retrieve information about clan's current clan war

        :param clanTag: Tag of the clan
        """
        r = Route("GET", API_PATHS["retrieve_current_clan_war"].format(clanTag=clanTag))
        return await self.request(r)

    async def get_clan_information(self, clanTag: str):
        """
        get_clan_information - Get information about a single clan by clan tag.

        :param clanTag: Tag of the clan
        """
        r = Route("GET", API_PATHS["get_clan_information"].format(clanTag=clanTag))
        return await self.request(r)

    async def get_clan_members(self, clanTag: str, limit: int = None, after: str = None, before: str = None):
        """
        get_clan_members - List clan members

        :param clanTag: Tag of the clan
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["list_clan_members"].format(clanTag=clanTag),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def get_current_river_race(self, clanTag: str):
        """
        get_current_river_race - Retrieve information about clan's current river race

        :param clanTag: Tag of the clan
        """
        r = Route("GET", API_PATHS["retrieve_current_clan_river_race"].format(clanTag=clanTag))
        return await self.request(r)

    # Player Information

    async def get_player_information(self, playerTag: str):
        """
        get_player_information - Get information about a single player by player tag. Player tags can be found either in game or by from clan member lists.

        :param playerTag: Tag of the player
        """
        r = Route("GET", API_PATHS["get_player_information"].format(playerTag=playerTag))
        return await self.request(r)

    async def get_player_upcoming_chests(self, playerTag: str):
        """
        get_player_upcoming_chests - Get list of reward chests that the player will receive next in the game.

        :param playerTag: Tag of the player
        """
        r = Route("GET", API_PATHS["get_upcoming_player_chests"].format(playerTag=playerTag))
        return await self.request(r)

    async def get_player_battle_log(self, playerTag: str):
        """
        get_player_battle_log - Get list of recent battles for a player.

        :param playerTag: Tag of the player
        """
        r = Route("GET", API_PATHS["get_player_battle_log"].format(playerTag=playerTag))
        return await self.request(r)

    # Cards

    async def get_available_cards(self, limit: int = None, after: str = None, before: str = None):
        """
        get_available_cards - Get list of available cards

        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["get_available_cards"],
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    # Tournaments
    async def search_tournaments(self, name: str = None, limit: int = None, after: str = None, before: str = None):
        """
        search_tournaments - Search all tournaments by name. It is not possible to specify ordering for results so clients should not rely on any
                            specific ordering as that may change in the future releases of the API.

        :param name: Search clans by name. If name is used as part of search query, it needs to be at least three characters long.
                     Name search parameter is interpreted as wild card search, so it may appear anywhere in the clan name.
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"name": name, "limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["search_tournaments"],
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def get_information_about_tournament(self, tournamentTag: str):
        """
        retrieve_information_about_tournament - Get information about a single tournament by a tournament tag.

        :param tournamentTag: Tag of the tournament to retrieve
        """
        r = Route(
            "GET",
            API_PATHS["get_tournament_information"].format(tournamentTag=tournamentTag),
        )
        return await self.request(r)

    # Locations

    async def get_clan_rankings_for_specific_location(
        self, locationId: int, limit: int = None, after: str = None, before: str = None
    ):
        """
        get_clan_rankings_for_specific_location - Get clan rankings for a specific location

        :param locationId: Identifier of the location to retrieve.
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["clan_ranking_for_specific_location"].format(locationId=locationId),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def get_player_rankings_for_specific_location(
        self, locationId: int, limit: int = None, after: str = None, before: str = None
    ):
        """
        get_player_rankings_for_specific_location - Get player rankings for a specific location

        :param locationId: Identifier of the location to retrieve.
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["player_ranking_for_specific_location"].format(locationId=locationId),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def get_clan_war_rankings_for_specific_location(
        self, locationId: int, limit: int = None, after: str = None, before: str = None
    ):
        """
        get_clan_war_rankings_for_specific_location - Get clan war rankings for a specific location

        :param locationId: Identifier of the location to retrieve.
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["clan_war_ranking_for_specific_location"].format(locationId=locationId),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    async def get_top_player_league_season(self, seasonId: str):
        """
        get_top_player_league_season - Get top player league season.

        :param seasonId: Id of player league season
        """
        r = Route("GET", API_PATHS["top_player_league_season"].format(seasonId=seasonId))
        return await self.request(r)

    async def get_top_player_rankings_for_season(self, seasonId: str):
        """
        get_top_player_rankings_for_season - Get top player rankings for a season.

        :param seasonId: Id of player league season
        """
        r = Route(
            "GET",
            API_PATHS["top_player_ranking_for_a_season"].format(seasonId=seasonId),
        )
        return await self.request(r)

    async def get_list_of_top_players_league_seasons(self):
        """
        get_list_of_top_players_league_seasons - Lists top player league seasons.
        """
        r = Route("GET", API_PATHS["list_of_top_player_league_seasons"])
        return await self.request(r)

    async def get_list_of_locations(self, limit: int = None, after: str = None, before: str = None):
        """
        get_list_of_locations - Lists locations

        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route("GET", API_PATHS["list_locations"], params=self.remove_none(query_params))
        return await self.request(r)

    async def get_information_on_location(self, locationId: str):
        """
        get_information_on_location - Get information about specific location

        :param locationId: Identifier of the location to retrieve.
        """
        r = Route("GET", API_PATHS["location_information"].format(locationId=locationId))
        return await self.request(r)

    async def get_global_tournament_rankings(
        self,
        tournamentTag: str,
        limit: int = None,
        after: str = None,
        before: str = None,
    ):
        """
        get_global_tournament_rankings - Get global tournament rankings

        :param tournamentTag: Tag of the tournament to retrieve
        :param limit: Limit the number of items returned in the response.
        :param after: Return only items that occur after this marker.
                      Before marker can be found from the response, inside the 'paging' property.
                      Note that only after or before can be specified for a request, not both.
        :param before: Return only items that occur before this marker.
                    Before marker can be found from the response, inside the 'paging' property.
                    Note that only after or before can be specified for a request, not both.
        """
        query_params = {"limit": limit, "after": after, "before": before}
        r = Route(
            "GET",
            API_PATHS["global_tournament_rankings"].format(tournamentTag=tournamentTag),
            params=self.remove_none(query_params),
        )
        return await self.request(r)

    # Global Tournaments

    async def get_list_of_global_tournaments(self):
        """
        get_list_of_global_tournaments - Get list of global tournaments
        """
        r = Route("GET", API_PATHS["list_of_global_tournaments"])
        return await self.request(r)
